package link.infra.packwiz.installer.request.handlers;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.function.Predicate;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

public abstract class RequestHandlerZip extends RequestHandlerHTTP {
	
	protected final boolean modeHasFolder;
	
	public RequestHandlerZip(boolean modeHasFolder) {
		this.modeHasFolder = modeHasFolder;
	}
	
	private String removeFolder(String name) {
		if (modeHasFolder) {
			return name.substring(name.indexOf("/")+1);
		} else {
			return name;
		}
	}
	
	private class ZipReader {
		
		private final ZipInputStream zis;
		private final Map<URI, byte[]> readFiles = new HashMap<URI, byte[]>();
		// Write lock implies access to ZipInputStream - only 1 thread must read at a time!
		final ReentrantLock filesLock = new ReentrantLock();
		private ZipEntry entry;

		public ZipReader(InputStream zip) {
			zis = new ZipInputStream(zip);
		}
		
		// File lock must be obtained before calling this function
		private byte[] readCurrFile() throws IOException {
			byte[] bytes = new byte[(int) entry.getSize()];
			DataInputStream dis = new DataInputStream(zis);
			dis.readFully(bytes);
			return bytes;
		}
		
		// File lock must be obtained before calling this function
		private byte[] findFile(URI loc) throws IOException, URISyntaxException {
			while (true) {
				entry = zis.getNextEntry();
				if (entry == null) {
					return null;
				}
				byte[] data = readCurrFile();
				URI fileLoc = new URI(removeFolder(entry.getName()));
				if (loc.equals(fileLoc)) {
					return data;
				} else {
					readFiles.put(fileLoc, data);
				}
			}
		}
		
		public InputStream getFileInputStream(URI loc) throws Exception {
			filesLock.lock();
			// Assume files are only read once, allow GC by removing
			byte[] file = readFiles.remove(loc);
			if (file != null) {
				filesLock.unlock();
				return new ByteArrayInputStream(file);
			}
			
			file = findFile(loc);
			filesLock.unlock();
			if (file != null) {
				return new ByteArrayInputStream(file);
			}
			return null;
		}
		
		public URI findInZip(Predicate<URI> matches) throws Exception {
			filesLock.lock();
			for (URI file : readFiles.keySet()) {
				if (matches.test(file)) {
					filesLock.unlock();
					return file;
				}
			}
			
			while (true) {
				entry = zis.getNextEntry();
				if (entry == null) {
					filesLock.unlock();
					return null;
				}
				byte[] data = readCurrFile();
				URI fileLoc = new URI(removeFolder(entry.getName()));
				readFiles.put(fileLoc, data);
				if (matches.test(fileLoc)) {
					filesLock.unlock();
					return fileLoc;
				}
			}
		}
		
	}
	
	private final Map<URI, ZipReader> cache = new HashMap<URI, ZipReader>();
	final ReentrantReadWriteLock cacheLock = new ReentrantReadWriteLock();
	
	protected abstract URI getZipUri(URI loc) throws Exception;
	
	protected abstract URI getLocationInZip(URI loc) throws Exception;
	
	@Override
	public abstract boolean matchesHandler(URI loc);

	@Override
	public InputStream getFileInputStream(URI loc) throws Exception {
		URI zipUri = getZipUri(loc);
		cacheLock.readLock().lock();
		ZipReader zr = cache.get(zipUri);
		cacheLock.readLock().unlock();
		if (zr == null) {
			cacheLock.writeLock().lock();
			// Recheck, because unlocking read lock allows another thread to modify it
			zr = cache.get(zipUri);
			if (zr == null) {
				InputStream str = super.getFileInputStream(zipUri);
				if (str == null) {
					cacheLock.writeLock().unlock();
					return null;
				}
				zr = new ZipReader(str);
				cache.put(zipUri, zr);
			}
			cacheLock.writeLock().unlock();
		}
		
		return zr.getFileInputStream(getLocationInZip(loc));
	}
	
	protected URI findInZip(URI loc, Predicate<URI> matches) throws Exception {
		URI zipUri = getZipUri(loc);
		cacheLock.readLock().lock();
		ZipReader zr = cache.get(zipUri);
		cacheLock.readLock().unlock();
		if (zr == null) {
			cacheLock.writeLock().lock();
			// Recheck, because unlocking read lock allows another thread to modify it
			zr = cache.get(zipUri);
			if (zr == null) {
				zr = new ZipReader(super.getFileInputStream(zipUri));
				cache.put(zipUri, zr);
			}
			cacheLock.writeLock().unlock();
		}
		
		return zr.findInZip(matches);
	}

}
